"""
Generic list document chain.
"""

import re
from urllib.parse import urlencode

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan import formatters
from mongoframes import ASC, DESC, And, In, InvalidPage, Or, Paginator, Q

from . import factories

__all__ = ['list_chains']


# Define the chains
list_chains = ChainMgr()

# GET
list_chains['get'] = Chain([
    'authenticate',
    'init_form',
    'validate',
    [
        [
            'search',
            'filter',
            'sort',
            'paginate',
            'form_info',
            'decorate',
            'render_template'
        ],
        ['redirect']
    ]
])


# Define the links
list_chains.set_link(factories.authenticate())
list_chains.set_link(factories.init_form('list', populate_on=['GET']))
list_chains.set_link(factories.validate('Invalid page or search request'))
list_chains.set_link(factories.decorate('list'))
list_chains.set_link(factories.render_template('list'))
list_chains.set_link(factories.redirect('list'))

@list_chains.link
def search(state):
    """
    Build a database query based on the `q` parameter within the request to
    filter the paginated documents.

    This link adds the `query` key to the state containing the database query.
    """
    state.query = None

    # If no query was provided then skip this step
    if 'q' not in state.form.data or not state.form.data['q']:
        return

    # Replace accented characters in the string with closest matching
    # equivalent non-accented characters.
    q = formatters.text.remove_accents(state.form.data['q'])

    # Make the query safe for regular expressions
    q = re.escape(q)

    # Build the query to search
    assert state.manage_config.list_search_fields, 'No search fields defined'

    q_match = re.compile(q, re.I)
    conditions = []
    for field in state.manage_config.list_search_fields:
        conditions.append(Q[field] == q_match)

    state.query = Or(*conditions)

@list_chains.link
def filter(state):
    """
    Apply any filters to the query.
    """

    # Check that there are filters to apply
    if not hasattr(state.form, 'filters'):
        return

    filter_conditions = []

    for filter_field in state.form.filters:

        # Get the name of the filter field against excluding the prefix
        based_field_name = filter_field.name.split('filters-', 1)[1]

        # Check for a custom filter handler
        custom_func_name = 'filter_' + based_field_name
        if hasattr(state.form.filters, custom_func_name):
            # Custom filter function exists so call it
            custom_func = getattr(state.form.filters, custom_func_name)
            condition = custom_func(state.form, filter_field)
            if condition:
                filter_conditions.append(condition)

        else:
            # No custom filter apply the default filter
            filter_value = filter_field.data

            # Build the conditions
            if filter_value:
                if isinstance(filter_value, list):
                    filter_conditions.append(
                            In(Q[based_field_name], filter_value))

                else:
                    filter_conditions.append(
                            Q[based_field_name] == filter_value)

    # Apply the filter conditions to the query
    if filter_conditions:
        if state.query:
            state.query = And(state.query, *filter_conditions)
        else:
            state.query = And(*filter_conditions)

@list_chains.link
def sort(state):
    """
    Build the sort operation to order the paginated documents.

    This link adds the `sort_by` key to the state containing the sort operation.
    """
    state.sort = None

    # If no sort field/direction was provided then skip this step
    if 'sort_by' not in state.form.data or not state.form.data['sort_by']:
        return

    # Build a sort operation from the form data
    state.sort_by = [(
        state.form.data['sort_by'].lstrip('-'),
        DESC if state.form.data['sort_by'].startswith('-') else ASC
        )]

@list_chains.link
def paginate(state):
    """
    Select a paginated list of documents from the database.

    This link adds `page` and `paginator` keys to the state containing the
    the paged results and the document paginator.
    """

    # Select the documents in paginated form
    state.paginator = Paginator(
        state.manage_config.frame_cls,
        state.query or {},
        sort=state.sort_by or None,
        per_page=state.manage_config.per_page,
        orphans=state.manage_config.orphans,
        projection=state.manage_config.list_projection
        )

    # Select the requested page
    try:
        state.page = state.paginator[state.form.data.get('page', 1)]
    except InvalidPage:
        return flask.redirect(flask.url_for(flask.request.url_rule.endpoint))

@list_chains.link
def form_info(state):
    """
    This link adds information about the form to the `state.form` instance
    (against `_info`, e.g `state.form._info`):

    - `filters_applied` flags that at least one field in the filters form has
      been applied.
    - `paging_qs` a URL that can be used to retain the form's state when
      navigating between pages.
    - `show_search_button` flags if there are any visible search fields in the
      form (excluding fields in the filters field form).

    NOTE: We store information against `_info` (prefixing wth the an underscore)
    because it prevents potential name clashes with fields that might be defined
    against the form, further for the most part filter forms are generated using
    a template macro which is aware of/expects this attribute and so we
    typcially don't expect developers to interact with this information
    directly.
    """

    if not state.form:
        return

    # Allocate storage for the form information against the form
    form_info = {}
    state.form._info = form_info

    # Check to see if any fields in the `filters` field form have been set
    form_info['filters_applied'] = False
    if hasattr(state.form, 'filters'):
        for field in state.form.filters:
            if field.data:
                form_info['filters_applied'] = True
                break

    # Generate a paging URL for the form
    paging_fields = [f for f in state.form if f.name not in {'filters', 'page'}]
    if hasattr(state.form, 'filters'):
        paging_fields += list(state.form.filters)

    params = {f.name: f.data for f in paging_fields if f.data}
    form_info['paging_query_string'] = urlencode(params, True)

    # Determine if there are any visible fields in the form that are not part
    # of pagination, sorting of the (advanced) filter field.
    form_info['show_search_button'] = False
    none_search_fields = {'filters', 'page', 'sort_by'}
    for field in state.form:
        if field.type != 'HiddenField' and field.name not in none_search_fields:
            form_info['show_search_button'] = True
            break


"""
Link factories for manhattan views.
"""


import bson
import os
import urllib

import flask
import inflection
from manhattan.nav import Nav, NavItem
from manhattan import secure

# Optional imports
try:
    from manhattan.assets import Asset
except ImportError as e:
    pass

__all__ = [
    'authenticate',
    'get_document',
    'init_form',
    'redirect',
    'render_template',
    'store_assets',
    'validate'
    ]


def authenticate(
        user_g_key='backend_user',
        sign_in_endpoint='backend_users.manage__sign_in',
        sign_out_endpoint='backend_users.manage__sign_out'
        ):
    """
    Authenticate the caller has permission to call the view.

    This link adds a `manage_user` key to the the state containing the currently
    signed in user.
    """

    def authenticate(state):
        # Get the signed in user
        state.manage_user = flask.g.get(user_g_key)

        # We're not allowed to access this view point so determine if that's
        # because we're not sign-in or because we don't have permission.
        if not state.manage_user:
            # We need to sign-in to view this endpoint

            # Forward the user to the sign-in page with the requested URL as the
            # `next` parameter.
            redirect_url = flask.url_for(
                sign_in_endpoint,
                next=secure.safe_redirect_url(
                    flask.request.url,
                    [flask.url_for(sign_out_endpoint)]
                    )
                )
            return flask.redirect(redirect_url)

        # Check if we're allowed to access this requested endpoint
        if not Nav.allowed(flask.request.endpoint, **flask.request.view_args):

            # We don't have permission to view this endpoint
            flask.abort(403, 'Permission denied')

    return authenticate

def decorate(view_type, title=None, last_breadcrumb=None, tabs=None):
    """
    Return a function that will add decor information to the state for the named
    view.

    Decor refers to content that is common to the majority of pages (for example
    the page title). By default decor includes:

    - `actions`
    - `breadcrumb`
    - `tabs`
    - `title`
    - `result_action` (only added to `list` view types).

    This link adds a `decor` key to the state.
    """

    def _get_breadcrumb_label(config, view_type):
        # Given a view type return the associated default breadcrumb label
        if view_type is 'list':
            return config.name_plural.capitalize()

        elif view_type is 'view':
            return config.name.capitalize() + ' details'

        return inflection.underscore(view_type).capitalize()

    def decorate(state):
        # Add `decor` to the state which all decorations will be assigned to
        state.decor = {}

        # Localize `manage_config`
        config = state.manage_config

        # Title
        state.decor['title'] = title
        if not title:
            if view_type is 'list':
                state.decor['title'] = config.name_plural.capitalize()

            elif view_type is 'view':
                state.decor['title'] = config.titleize(state[config.var_name])

            elif view_type is 'add':
                state.decor['title'] = 'Add ' + config.name

            else:
                state.decor['title'] = '{view_type} {document}'.format(
                    document=config.titleize(state[config.var_name]),
                    view_type=inflection.underscore(view_type).capitalize()
                    )

        # Breadcrumb
        breadcrumbs = Nav.local_menu()

        # All view types except list are assumed to start their breadcrumb trail
        # with list...
        if view_type is not 'list':
            if Nav.exists(config.get_endpoint('list')):
                list_item = NavItem(
                    _get_breadcrumb_label(config, 'list'),
                    config.get_endpoint('list')
                    )
                breadcrumbs.add(list_item)

        # ...followed by view (with the exception of add and view).
        if view_type not in ['add', 'list', 'view']:
            if Nav.exists(config.get_endpoint('view')):
                _id = state[config.var_name]._id
                view_args = dict([(config.var_name, _id)])
                view_item = NavItem(
                    _get_breadcrumb_label(config, 'view'),
                    config.get_endpoint('view'),
                    view_args=view_args
                    )
                breadcrumbs.add(view_item)

        # Add the last breadcrumb
        item = NavItem(
            last_breadcrumb or _get_breadcrumb_label(config, view_type),
            ''
            )
        breadcrumbs.add(item)

        # Set the breadcrumbs against the state
        state.decor['breadcrumbs'] = breadcrumbs

        # Actions
        actions = Nav.local_menu()

        # List pages are assumed to support add actions
        if view_type is 'list':
            if Nav.exists(config.get_endpoint('add')):
                actions.add(NavItem('Add', config.get_endpoint('add')))

        # View pages are assumed to support update and delete actions
        elif view_type is 'view':
            _id = state[config.var_name]._id
            view_args = dict([(config.var_name, _id)])

            if Nav.exists(config.get_endpoint('update')):
                endpoint = config.get_endpoint('update')
                actions.add(NavItem('Update', endpoint, view_args=view_args))

            if Nav.exists(config.get_endpoint('delete')):
                endpoint = config.get_endpoint('delete')
                actions.add(NavItem('Delete', endpoint, view_args=view_args))

        # Update pages are assumed to support delete actions if there is no view
        # page.
        elif view_type is 'update' \
                and not Nav.exists(config.get_endpoint('view')):

            if Nav.exists(config.get_endpoint('delete')):
                _id = state[config.var_name]._id
                endpoint = config.get_endpoint('delete')
                view_args = dict([(config.var_name, _id)])
                actions.add(NavItem('Delete', endpoint, view_args=view_args))

        # Set the actions against the state
        state.decor['actions'] = actions

        # Tabs (if a tabs menu has been specified we attempt to find it and
        # include it in the decor).
        state.decor['tabs'] = None
        if tabs:
            state.decor['tabs'] = Nav.menu(tabs, raise_unless_exists=True)

        # View link

        # List view types are provided with a result action, the result action
        # should be a function that accepts typically a single document as an
        # argument and returns a `Nav.query` result.
        if view_type is 'list':

            def results_action(document):
                # A generic results action
                args = {}
                args[config.var_name] = document._id

                # Check for view link
                link = Nav.query(config.get_endpoint('view'), **args)
                if not link.exists:

                    # Check for update link
                    link = Nav.query(config.get_endpoint('update'), **args)
                    if link.exists:
                        return link

                return link

            state.decor['results_action'] = results_action

    return decorate

def get_document(projection=None):
    """
    Return a function that will attempt to retreive a document from the
    database by `_id` using the `var_name` named parameter in the request.

    This link adds a `{state.manage_config.var_name}` key to the the state
    containing the document retreived.

    Optionally a projection to use when getting the document can be specified,
    this can either be a dictionary or the name of the projection as defined
    against the config, e.g: `get_document('view')` will use the projection
    defined against the config attribute `view_projection`.
    """

    def get_document(state):
        # Get the Id of the document passed in the request
        document_id = flask.request.values.get(state.manage_config.var_name)

        # Attempt to convert the Id to the required type
        try:
            document_id = bson.objectid.ObjectId(document_id)
        except bson.errors.InvalidId:
            flask.abort(404)

        # If there's a projection defined check if it's a string, if so we need
        # to look for the projection against the config.
        _projection = None
        if projection:
            if isinstance(projection, str):
                if hasattr(state.manage_config, projection + '_projection'):
                    _projection = getattr(
                        state.manage_config,
                        projection + '_projection'
                        )
            elif isinstance(projection, dict):
                _projection = projection

        # Attempt to retrieve the document
        kwargs = {}
        if _projection:
            kwargs['projection'] = _projection
        document = state.manage_config.frame_cls.by_id(document_id, **kwargs)

        if not document:
            flask.abort(404)

        # Set the document against the state
        state[state.manage_config.var_name] = document

    return get_document

def init_form(view_name, populate_on=None):
    """
    Return a function that will initialize a form for the named generic view
    (e.g list, add, update, etc.)

    Optionally the CSRF projection for the form can be enabled (default) or
    disabled.

    This link adds a `form` key to the the state containing the initialized
    form.
    """

    # If populate_on is not specified then we default to `POST`
    if populate_on is None:
        populate_on = ['POST']

    def init_form(state):
        # Get the form class
        form_cls = state.manage_config.get_form_cls(view_name)
        assert form_cls, 'No form defined for this view {0}'.format(view_name)

        # Initialize the form
        formdata = None
        if flask.request.method in populate_on:
            if flask.request.method in ['POST', 'PUT']:
                formdata = flask.request.form
            else:
                formdata = flask.request.args

        # If a document is assign to the state then is is sent as the first
        # argument when initializing the form.
        obj = None
        if state.manage_config.var_name in state:
            obj = state[state.manage_config.var_name]
        state.form = form_cls(formdata, obj=obj)

    return init_form

def redirect(endpoint, include_id=False):
    """
    Return a function that will trigger a redirect to the given endpoint.

    Optionally an Id for the current document in the state can be added to the
    URL, e.g `url_for('view.user', user=user._id)` by passing `include_id` as
    True.
    """

    def redirect(state):
        # Build the URL arguments
        url_args = {}
        if include_id:
            url_args[state.manage_config.var_name] = \
                    state[state.manage_config.var_name]._id

        # Get the URL for the endpoint
        prefix = state.manage_config.endpoint_prefix
        if state.manage_config.endpoint_prefix:
            url = flask.url_for('.' + prefix + endpoint, **url_args)
        else:
            url = flask.url_for('.' + endpoint, **url_args)

        # Return the redirect response
        return flask.redirect(url)

    return redirect

def render_template(template_name):
    """
    Return a function that will render the named template. The state object is
    used as template arguments.
    """

    def render_template(state):
        # Build the template filepath
        template_filepath = os.path.join(
            state.manage_config.template_path,
            template_name + '.html'
            )

        # Render the template
        return flask.render_template(template_filepath, **state)

    return render_template

def store_assets():
    """
    Return a function that will convert temporary assets to permenant assets.
    """

    def store_assets(state):

        # Check that the app supports an asset manager, if not then there's
        # nothing to do.
        if not hasattr(flask.current_app, 'asset_mgr'):
            return
        asset_mgr = flask.current_app.asset_mgr

        # Get the document being added or updated
        document = state[state.manage_config.var_name]

        # Find any values which against the document which is a temporary asset
        update_fields = []
        for field in document.get_fields():
            value = document.get(field)

            # Ignore any value that's not a temporary asset
            if not (isinstance(value, Asset) and value.temporary):
                continue

            # Store the asset permenantly
            flask.current_app.asset_mgr.store(value)

            # Check if any variations are defined for the field
            if hasattr(state.manage_config, field + '_variations'):
                variations = getattr(state.manage_config, field + '_variations')

                # Store variations for the asset
                asset_mgr.generate_variations(value, variations)

            # Flag the field requires updating against the database
            update_fields.append(field)

        if update_fields:
            document.update(*update_fields)

    return store_assets

def validate(error_msg='Please review your submission'):
    """
    Return a function that will call validate against `state.form`. If the form
    is valid the function will return `True` or `False` if not.

    Optionally an `error_msg` can be passed, if the form fails to validate this
    will be flashed to the user.
    """

    def validate(state):
        assert state.form, 'No form to validate against'

        if state.form.validate():
            return True

        flask.flash(error_msg, 'error')
        return False

    return validate



"""
Generic add document chain.
"""

import flask
from manhattan.chains import Chain, ChainMgr

from . import factories

__all__ = ['add_chains']


# Define the chains
add_chains = ChainMgr()

# GET
add_chains['get'] = Chain([
    'authenticate',
    'init_form',
    'decorate',
    'render_template'
])

# POST
add_chains['post'] = Chain([
    'authenticate',
    'init_form',
    'validate',
    [
        [
            'init_document',
            'add_document',
            'store_assets',
            'redirect'
        ],
        [
            'decorate',
            'render_template'
        ]
    ]
])


# Define the links
add_chains.set_link(factories.authenticate())
add_chains.set_link(factories.init_form('add'))
add_chains.set_link(factories.validate())
add_chains.set_link(factories.store_assets())
add_chains.set_link(factories.decorate('add'))
add_chains.set_link(factories.render_template('add'))
add_chains.set_link(factories.redirect('view', include_id=True))

@add_chains.link
def init_document(state):
    """
    Initialize a new document and populates it using the submitted form data.

    This link adds a `{state.manage_config.var_name}` key to the the state
    containing the initialized document.
    """

    # Initialize a new document
    document = state.manage_config.frame_cls()

    # Populate the document from the form
    state.form.populate_obj(document)

    # Set the document against the state
    state[state.manage_config.var_name] = document

@add_chains.link
def add_document(state):
    """Add a document"""

    # Get the initialized document
    document = state[state.manage_config.var_name]

    assert document, \
            'No `{0}` set in state'.format(state.manage_config.var_name)

    # Check to see if the frame class supports `logged_insert`s and if so
    if hasattr(state.manage_config.frame_cls, 'logged_insert'):
        # Supports `logged_insert`
        document.logged_insert(state.manage_user)

    else:
        # Doesn't support `logged_insert`
        document.insert()

    assert document, \
            'No `{0}` set in state'.format(state.manage_config.var_name)

    # Flash message that the document was added
    flask.flash('{document} added.'.format(document=document))



"""
Generic delete document chain.
"""

import flask
from manhattan.chains import Chain, ChainMgr

from . import factories

__all__ = ['delete_chains']


# Define the chains
delete_chains = ChainMgr()

# GET
delete_chains['get'] = Chain([
    'authenticate',
    'get_document',
    'decorate',
    'render_template'
    ])

# POST
delete_chains['post'] = Chain([
    'authenticate',
    'get_document',
    'delete_document',
    'redirect'
    ])


# Define the links
delete_chains.set_link(factories.authenticate())
delete_chains.set_link(factories.get_document('delete'))
delete_chains.set_link(factories.decorate('delete'))
delete_chains.set_link(factories.render_template('delete'))
delete_chains.set_link(factories.redirect('list'))

@delete_chains.link
def delete_document(state):
    """Delete a document"""

    # Get the document
    document = state[state.manage_config.var_name]

    assert document, \
            'No `{0}` set in state'.format(state.manage_config.var_name)

    # Check to see if the frame class supports `logged_delete`s and if so
    if hasattr(state.manage_config.frame_cls, 'logged_delete'):
        # Supports `logged_delete`
        document.logged_delete(state.manage_user)

    else:
        # Doesn't support `logged_delete`
        document.delete()

    # Flash message that the document was deleted
    flask.flash('{document} deleted.'.format(document=document))


import flask
import jinja2

from manhattan.assets import Asset
from manhattan.manage.views.assets.upload import upload_chains as upload


class Manage:
    """
    The `Manage` class provides the initialization code for the package.
    """

    def __init__(self, app):
        self._app = app

        # Set-up a hidden blueprint to provide access to the templates
        blueprint = flask.Blueprint(
            '_manhattan_manage',
            __name__,
            template_folder='templates'
            )
        self._app.register_blueprint(blueprint)

    def manage_assets(self, root, backend, settings=None, path='/cms/'):
            """Set up support for managing assets"""

            # Configure the URL root for assets
            Asset._asset_root = root

            # Set up the asset manager
            self._app.asset_mgr = backend.AssetMgr(**(settings or {}))

            # Set up views for managing assets

            # Upload
            self._app.add_url_rule(
                path + 'upload-asset',
                endpoint='manage__upload_asset',
                view_func=upload.copy().flask_view(),
                methods=['POST']
                )

            # @@ Other views to come...


"""
Generic view document chain.
"""

from manhattan.chains import Chain, ChainMgr

from . import factories

__all__ = ['view_chains']


# Define the chains
view_chains = ChainMgr()

# GET
view_chains['get'] = Chain([
    'authenticate',
    'get_document',
    'decorate',
    'render_template'
    ])


# Define the links
view_chains.set_link(factories.authenticate())
view_chains.set_link(factories.get_document('view'))
view_chains.set_link(factories.decorate('view'))
view_chains.set_link(factories.render_template('view'))


"""
Utilities functions for manhattan views.
"""

import flask

__all__ = [
    'json_fail',
    'json_success'
    ]


def json_fail(reason, issues=None):
    """Return a fail response"""
    response = {'status': 'fail', 'payload': {'reason': reason}}
    if issues:
        response['payload']['issues'] = issues
    return flask.jsonify(response)

def json_success(payload=None):
    """Return a success response"""
    response = {'status': 'success'}
    if payload:
        response['payload'] = payload
    return flask.jsonify(response)


"""
Generic typeahead document chain.

The generic typeahead view is designed to return *all* relevant results for a
user query and leaves the final filtering and sorting of the results to the
client-side caller.
"""

import re

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan import formatters
from mongoframes import ASC, DESC, InvalidPage, Or, Paginator, Q

from . import factories
from .utils import json_fail, json_success

__all__ = ['typeahead_chains']


# Define the chains
typeahead_chains = ChainMgr()

# GET
typeahead_chains['get'] = Chain([
    'authenticate',
    'search',
    'select',
    'render_json'
])


# Define the links
typeahead_chains.set_link(factories.authenticate())

@typeahead_chains.link
def search(state):
    """
    Build a database query based on the `q` parameter within the request to
    filter the typeahead results.

    This link adds the `query` key to the state containing the database query.
    """
    state.query = None

    # If the query isn't long enough to query the database then return a failed
    # response.
    q = flask.request.args.get('q', '').strip()
    if len(q) < state.manage_config.typeahead_match_len:
        return json_fail(
            'The query string is too short, it should be at least {0} '
            'characters long'.format(state.manage_config.typeahead_match_len)
            )

    # Replace accented characters in the string with closest matching
    # equivalent non-accented characters.
    q = formatters.text.remove_accents(q)

    # Shorten the query string to the match length
    q = q[:state.manage_config.typeahead_match_len]

    # Make the query safe for regular expressions
    q = re.escape(q)

    # Build the query to search
    typeahead_field = state.manage_config.typeahead_field

    assert typeahead_field, 'No typeahead field defined'

    if state.manage_config.typeahead_match_type == 'contains':
        state.query = Q[typeahead_field] == re.compile(q)
    elif state.manage_config.typeahead_match_type == 'startswith':
        state.query = Q[typeahead_field] == re.compile(r'^' + q)

    assert state.query, "'{0}' is not a valid typeahead match type"

@typeahead_chains.link
def select(state):
    """
    Select the documents for the typeahead.

    This link adds the `{state.manage_config.var_name_plural}` key to the state
    containing the selected database documents.
    """

    # Select the documents
    documents = state.manage_config.frame_cls.many(
        state.query,
        projection=state.manage_config.typeahead_projection
        )

    state[state.manage_config.var_name_plural] = documents

@typeahead_chains.link
def render_json(state):
    """
    Convert the selected documents to a JSON result for the client-side
    typeahead.
    """

    # Convert the matching documents into a list of items for the typeahead
    items = []
    for document in state[state.manage_config.var_name_plural]:
        item = {}

        # Id
        if state.manage_config.typeahead_id_attr:
            item['id'] = getattr(
                document,
                state.manage_config.typeahead_id_attr
                )
        else:
            item['id'] = str(document._id)

        # Value
        item['value'] = getattr(
            document,
            state.manage_config.typeahead_field
            )

        # Label
        if state.manage_config.typeahead_label_attr:
            item['label'] = getattr(
                document,
                state.manage_config.typeahead_label_attr
                )
        else:
            item['label'] = str(document)

        items.append(item)

    return json_success({'items': items})


"""
Generic update document chain.
"""

import flask
from manhattan.chains import Chain, ChainMgr

from . import factories

__all__ = ['update_chains']


# Define the chains
update_chains = ChainMgr()

# GET
update_chains['get'] = Chain([
    'authenticate',
    'get_document',
    'init_form',
    'decorate',
    'render_template'
    ])

# POST
update_chains['post'] = Chain([
    'authenticate',
    'get_document',
    'init_form',
    'validate',
    [
        [
            'build_form_data',
            'update_document',
            'store_assets',
            'redirect'
        ],
        [
            'decorate',
            'render_template'
        ]
    ]
])

# Define the links
update_chains.set_link(factories.authenticate())
update_chains.set_link(factories.get_document('update'))
update_chains.set_link(factories.init_form('update'))
update_chains.set_link(factories.validate())
update_chains.set_link(factories.store_assets())
update_chains.set_link(factories.decorate('update'))
update_chains.set_link(factories.render_template('update'))
update_chains.set_link(factories.redirect('view', include_id=True))

@update_chains.link
def build_form_data(state):
    """
    Generate the form data that will be used to update the document.

    This link adds a `form_data` key to the the state containing the initialized
    form.
    """
    state.form_data = state.form.data

@update_chains.link
def update_document(state):
    """Update a document"""

    # Get the initialized document
    document = state[state.manage_config.var_name]

    assert document, \
            'No `{0}` set in state'.format(state.manage_config.var_name)

    # Check to see if the frame class supports `logged_update`s and if so
    if hasattr(state.manage_config.frame_cls, 'logged_update'):
        # Supports `logged_update`
        document.logged_update(state.manage_user, state.form_data)

    else:
        # Doesn't support `logged_update`
        for k, v in state.form_data.items():
            setattr(document, k, v)
        document.update()

    assert document, \
            'No `{0}` set in state'.format(state.manage_config.var_name)

    # Flash message that the document was updated
    flask.flash('{document} updated.'.format(document=document))
